/*	TOP-SURF: a visual words toolkit
	Copyright (C) 2010 LIACS Media Lab, Leiden University,
	                   Bart Thomee (bthomee@liacs.nl),
					   Erwin M. Bakker (erwin@liacs.nl)	and
					   Michael S. Lew (mlew@liacs.nl).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    See http://www.gnu.org/licenses/gpl.html for the full license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	In addition, this work is covered under the Creative Commons
	Attribution license version 3.
    See http://creativecommons.org/licenses/by/3.0/ for the full license.
*/

#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

#if !defined WIN32 && !defined _WIN32 && !defined WIN64 && !defined _WIN64
#include <libgen.h>
#endif

#include "topsurf/topsurf.h"
#include "topsurf/opencv/cv.h"
#include "topsurf/config.h" // contains some useful functions such as safe string printing to the console

// print out usage
int Usage(const char *exe)
{
	string filename;
#if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
	char szFile[_MAX_PATH], szExt[_MAX_EXT];
	_splitpath(exe, NULL, NULL, szFile, szExt);
	filename = szFile;
	filename += szExt;
#else
	char *path = strdup(exe);
	filename = "./";
	filename += (string)basename(path);
	free(path);
#endif
	
	SAFE_FLUSHPRINT(stdout, "\n");
	SAFE_FLUSHPRINT(stdout, "=====================\n");
	SAFE_FLUSHPRINT(stdout, "======= Usage =======\n");
	SAFE_FLUSHPRINT(stdout, "=====================\n");
	SAFE_FLUSHPRINT(stdout, "\n");
	SAFE_FLUSHPRINT(stdout, "To extract the descriptors of an image:\n");
	SAFE_FLUSHPRINT(stdout, "  %s %s %s %s %s %s\n", filename.c_str(), "extract", "[dictdir]", "[img resize]", "[top]", "[img]" );
	SAFE_FLUSHPRINT(stdout, "  e.g. %s %s %s %s %s %s\n", filename.c_str(), "extract", "D:\\dictionary", "256", "100", "image.jpg");
	SAFE_FLUSHPRINT(stdout, "The extracted descriptor will be saved with the same filename as the image,\n"
		"but with the .top extension.\n");
	SAFE_FLUSHPRINT(stdout, "\n");
	SAFE_FLUSHPRINT(stdout, "To compare two descriptors:\n");
	SAFE_FLUSHPRINT(stdout, "  %s %s %s %s %s\n", filename.c_str(), "compare", "[descriptor1]", "[descriptor2]", "[mode]");
	SAFE_FLUSHPRINT(stdout, "  where mode is either 'cosine' or 'absolute'.\n");
	SAFE_FLUSHPRINT(stdout, "  e.g. %s %s %s %s %s\n", filename.c_str(), "compare", "image1.top", "image2.top", "cosine");
	SAFE_FLUSHPRINT(stdout, "\n");
	SAFE_FLUSHPRINT(stdout, "To visualize the locations of the detected visual words in an image:\n");
	SAFE_FLUSHPRINT(stdout, "  %s %s %s %s %s\n", filename.c_str(), "visualize", "[image]", "[descriptor]", "[output]");
	SAFE_FLUSHPRINT(stdout, "  e.g. %s %s %s %s %s\n", filename.c_str(), "visualize", "image.jpg", "image.top", "output.jpg");
	SAFE_FLUSHPRINT(stdout, "\n");
	SAFE_FLUSHPRINT(stdout, "To save the locations of the detected visual words in an image to a text file:\n");
	SAFE_FLUSHPRINT(stdout, "  %s %s %s %s\n", filename.c_str(), "locations", "[descriptor]", "[output]");
	SAFE_FLUSHPRINT(stdout, "  e.g. %s %s %s %s\n", filename.c_str(), "locations", "image.top", "output.txt");
	SAFE_FLUSHPRINT(stdout, "\n");
	SAFE_FLUSHPRINT(stdout, "Please refer to the README for more detailed information.\n");
	SAFE_FLUSHPRINT(stdout, "Press any key to continue...\n");
	_getch();
	return -1;
}

int main(int argc, char *argv[])
{
	// parse the command line
	// Note: first argument is always the path to the current executable
	// Note: see Usage function for instructions
	if (argc == 0)
		return -1;
	if (argc == 1)
		return Usage(argv[0]);

	// for troubleshooting, print out arguments
	for (int i = 0; i < argc; i++)
		SAFE_FLUSHPRINT(stdout, "arg%i: %s\n", i, argv[i]);
	
	// perform the requested functionality
	if (strcmp(argv[1], "extract") == 0)
	{
		if (argc != 6)
			return Usage(argv[0]);
		const char *dictionarydir = argv[2];
		int imagedim = atoi(argv[3]);
		int top = atoi(argv[4]);
		const char *imagename = argv[5];
		if (strlen(dictionarydir) == 0 || imagedim <= 0 || top <= 0 || strlen(imagename) == 0)
			return Usage(argv[0]);
		// load the image
		IplImage *image = cvLoadImage(imagename, CV_LOAD_IMAGE_COLOR);
		if (image == NULL)
		{
			SAFE_FLUSHPRINT(stderr, "could not load image %s\n", imagename);
			return -1;
		}
		// initialize topsurf
		SAFE_FLUSHPRINT(stdout, "Please wait while the TOP-SURF library initializes...\n");
		TopSurf topsurf(imagedim, top);
		if (!topsurf.LoadDictionary(dictionarydir))
		{
			cvReleaseImage(&image);
			return -1;
		}
		// extract the descriptor
		SAFE_FLUSHPRINT(stdout, "Extracting descriptor...\n");
		TOPSURF_DESCRIPTOR td;
		if (!topsurf.ExtractDescriptor(*image, td))
		{
			cvReleaseImage(&image);
			return -1;
		}
		// save the descriptor
		// Note: this only requires static access to the topsurf library, which
		//       is why we already terminated the dll above
		SAFE_FLUSHPRINT(stdout, "Saving descriptor...\n");
		char descname[MAX_PATH];
#if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
		char szDrive[_MAX_DRIVE], szDir[_MAX_PATH], szFile[_MAX_PATH];
		_splitpath(imagename, szDrive, szDir, szFile, NULL);
		SAFE_SPRINTF(descname, sizeof(descname), "%s%s%s%s", szDrive != NULL ? szDrive : "", szDir != NULL ? szDir : "", szFile, ".top");
#else
		char *path = strdup(imagename);
		string dir = (string)dirname(path);
		string filename = (string)basename(path);
		free(path);
		size_t pos = filename.find_last_of('.');
		filename = filename.substr(0, pos);
		SAFE_SPRINTF(descname, sizeof(descname), "%s%s%s%s", dir.c_str(), PATH_SEPARATOR_STRING, filename.c_str(), ".top");
#endif
		if (!TopSurf::SaveDescriptor(descname, td))
		{
			TopSurf::ReleaseDescriptor(td);
			cvReleaseImage(&image);
			return -1;
		}
		// release resources
		TopSurf::ReleaseDescriptor(td);
		cvReleaseImage(&image);
		return 0;
	}
	else if (strcmp(argv[1], "compare") == 0)
	{
		if (argc != 5)
			return Usage(argv[0]);
		const char *descname1 = argv[2];
		const char *descname2 = argv[3];
		const char *mode = argv[4];
		if (strlen(descname1) == 0 || strlen(descname2) == 0 || strlen(mode) == 0)
			return Usage(argv[0]);
		if (strcmp(mode, "cosine") == 0)
		{
			SAFE_FLUSHPRINT(stdout, "Comparing descriptors using cosine normalized difference...\n");
			TOPSURF_DESCRIPTOR td1, td2;
			if (!TopSurf::LoadDescriptor(descname1, td1) || !TopSurf::LoadDescriptor(descname2, td2))
			{
				TopSurf::ReleaseDescriptor(td1);
				TopSurf::ReleaseDescriptor(td2);
				return -1;
			}
			float distance = TopSurf::CompareDescriptorsCosine(td1, td2);
			TopSurf::ReleaseDescriptor(td1);
			TopSurf::ReleaseDescriptor(td2);
			SAFE_FLUSHPRINT(stdout, "The distance between both descriptors is %.2f...\n", distance);
			return 0;
		}
		else if (strcmp(mode, "absolute") == 0)
		{
			SAFE_FLUSHPRINT(stdout, "Comparing descriptors using absolute difference...\n");
			TOPSURF_DESCRIPTOR td1, td2;
			if (!TopSurf::LoadDescriptor(descname1, td1) || !TopSurf::LoadDescriptor(descname2, td2))
			{
				TopSurf::ReleaseDescriptor(td1);
				TopSurf::ReleaseDescriptor(td2);
				return -1;
			}
			float distance = TopSurf::CompareDescriptorsAbsolute(td1, td2);
			TopSurf::ReleaseDescriptor(td1);
			TopSurf::ReleaseDescriptor(td2);
			SAFE_FLUSHPRINT(stdout, "The distance between both descriptors is %.2f...\n", distance);
			return 0;
		}
	}
	else if (strcmp(argv[1], "visualize") == 0)
	{
		if (argc != 5)
			return Usage(argv[0]);
		const char *imagename1 = argv[2];
		const char *descname = argv[3];
		const char *imagename2 = argv[4];

		if (strlen(imagename1) == 0 || strlen(descname) == 0 || strlen(imagename2) == 0)
			return Usage(argv[0]);
		// load the image
		IplImage *image = cvLoadImage(imagename1, CV_LOAD_IMAGE_COLOR);
		if (image == NULL)
		{
			SAFE_FLUSHPRINT(stderr, "could not load image %s\n", imagename1);
			return -1;
		}
		// load the descriptor
		TOPSURF_DESCRIPTOR td;
		if (!TopSurf::LoadDescriptor(descname, td))
		{
			cvReleaseImage(&image);
			return -1;
		}
		// draw the points
		TopSurf::VisualizeDescriptor(*image, td);
		// release resources
		TopSurf::ReleaseDescriptor(td);
		// save the image
		if (!cvSaveImage(imagename2, image))
		{
			SAFE_FLUSHPRINT(stdout, "Could not save image %s\n", imagename2);
			cvReleaseImage(&image);
			return -1;
		}
		// release resources
		cvReleaseImage(&image);
		return 0;
	}
	else if (strcmp(argv[1], "locations") == 0)
	{
		if (argc != 4)
			return Usage(argv[0]);
		const char *descname = argv[2];
		const char *textname = argv[3];
		if (strlen(descname) == 0 || strlen(textname) == 0)
			return Usage(argv[0]);
		// load the descriptor
		TOPSURF_DESCRIPTOR td;
		if (!TopSurf::LoadDescriptor(descname, td))
			return -1;
		// open the output file
		FILE *file = fopen(textname, "w");
		if (file == NULL)
		{
			TopSurf::ReleaseDescriptor(td);
			SAFE_FLUSHPRINT(stdout, "Could not open %s for read\n", textname);
			return -1;
		}
		// print out the locations
		SAFE_FLUSHPRINT(file, "visual word\t          x\t          y\t      scale\torientation\n");
		for (int i = 0; i < td.count; i++)
		{
			// paint all points
			for (int j = 0; j < td.visualword[i].count; j++)
			{
				TOPSURF_LOCATION l = td.visualword[i].location[j];
				SAFE_FLUSHPRINT(file, "%11i\t%11.2f\t%11.2f\t%11.2f\t%11.2f\n", td.visualword[i].identifier, l.x, l.y, l.scale, l.orientation);
			}
		}
		fclose(file);
		// release resources
		TopSurf::ReleaseDescriptor(td);
		return 0;
	}
	return Usage(argv[0]);
}